@@ -1,5 +1,3 @@
-import struct
-
 from datetime import time as Time
 
 import pytest
@@ -214,12 +212,6 @@ def test_closed_connection(con):
         con.run("VALUES ('hw1'::text)")
 
 
-def test_network_error_on_connect(db_kwargs, mocker):
-    mocker.patch("pg8000.core.ci_unpack", side_effect=struct.error())
-    with pytest.raises(InterfaceError, match="network error"):
-        Connection(**db_kwargs)
-
-
 def test_version():
     try:
         from importlib.metadata import version
