@@ -12,6 +12,7 @@ from decimal import Decimal
 from enum import Enum
 from ipaddress import IPv4Address, IPv4Network
 from json import dumps
+from locale import LC_ALL, localeconv, setlocale
 from uuid import UUID
 
 import pytest
@@ -38,6 +39,7 @@ from pg8000.converters import (
     NUMERIC_ARRAY,
     PGInterval,
     POINT,
+    Range,
     SMALLINT_ARRAY,
     TIME,
     TIMESTAMP,
@@ -49,7 +51,6 @@ from pg8000.converters import (
     XID,
     pg_interval_in,
     pg_interval_out,
-    time_in,
 )
 
 
@@ -296,6 +297,28 @@ def test_timestamp_mismatch(con):
         con.run("SET SESSION TIME ZONE DEFAULT")
 
 
+@pytest.mark.parametrize(
+    "select,expected",
+    [
+        ["CAST('t' AS bool)", True],
+        ["5000::smallint", 5000],
+        ["5000::numeric", Decimal("5000")],
+        ["50.34::numeric", Decimal("50.34")],
+        ["5000::integer", 5000],
+        ["50000000000000::bigint", 50000000000000],
+        ["1.1::real", 1.1],
+        ["1.1::double precision", 1.1000000000000001],
+        ["'hello'::varchar(20)", "hello"],
+        ["'hello'::char(20)", "hello               "],
+        ["'hello'::text", "hello"],
+        ["(1,2)", ("1", "2")],
+    ],
+)
+def test_in(con, select, expected):
+    retval = con.run(f"SELECT {select}")
+    assert retval[0][0] == expected
+
+
 def test_name_out(con):
     # select a field that is of "name" type:
     con.run("SELECT usename FROM pg_user")
@@ -307,57 +330,6 @@ def test_oid_out(con):
     # It is sufficient that no errors were encountered.
 
 
-def test_boolean_in(con):
-    retval = con.run("SELECT cast('t' as bool)")
-    assert retval[0][0]
-
-
-def test_numeric_out(con):
-    for num in ("5000", "50.34"):
-        retval = con.run("SELECT " + num + "::numeric")
-        assert str(retval[0][0]) == num
-
-
-def test_int2_out(con):
-    retval = con.run("SELECT 5000::smallint")
-    assert retval[0][0] == 5000
-
-
-def test_int4_out(con):
-    retval = con.run("SELECT 5000::integer")
-    assert retval[0][0] == 5000
-
-
-def test_int8_out(con):
-    retval = con.run("SELECT 50000000000000::bigint")
-    assert retval[0][0] == 50000000000000
-
-
-def test_float4_out(con):
-    retval = con.run("SELECT 1.1::real")
-    assert retval[0][0] == 1.1
-
-
-def test_float8_out(con):
-    retval = con.run("SELECT 1.1::double precision")
-    assert retval[0][0] == 1.1000000000000001
-
-
-def test_varchar_out(con):
-    retval = con.run("SELECT 'hello'::varchar(20)")
-    assert retval[0][0] == "hello"
-
-
-def test_char_out(con):
-    retval = con.run("SELECT 'hello'::char(20)")
-    assert retval[0][0] == "hello               "
-
-
-def test_text_out(con):
-    retval = con.run("SELECT 'hello'::text")
-    assert retval[0][0] == "hello"
-
-
 def test_pg_interval_in(con):
     con.register_in_adapter(1186, pg_interval_in)
     retval = con.run(
@@ -517,20 +489,11 @@ def test_float8_array_out(con):
     assert f3 == [[[1, 2], [3, 4]], [[None, 6], [7, 8]]]
 
 
-CURRENCIES = {
-    "en_GB.UTF-8": "£",
-    "C.UTF-8": "$",
-    "C.UTF8": "$",
-}
-
 # Find the currency string
-for LC in ("LC_CTYPE", "LANG"):
-    try:
-        LC_VAL = os.environ[LC]
-        CURRENCY = CURRENCIES[LC_VAL]
-        break
-    except KeyError:
-        pass
+setlocale(LC_ALL, "")
+CURRENCY = localeconv()["currency_symbol"]
+if CURRENCY == "":
+    CURRENCY = "$"
 
 
 @pytest.mark.parametrize(
@@ -579,13 +542,12 @@ for LC in ("LC_CTYPE", "LANG"):
         [IPv4Address("192.168.0.1"), INET],  # inet
         [86722, XID],  # xid
         ["infinity", TIMESTAMP],  # timestamp
-        ["(2.3,1)", POINT],  # point
+        [(2.3, 1), POINT],  # point
         [{"name": "Apollo 11 Cave", "zebra": True, "age": 26.003}, JSON],  # json
         [{"name": "Apollo 11 Cave", "zebra": True, "age": 26.003}, JSONB],  # jsonb
     ],
 )
 def test_roundtrip_oid(con, test_input, oid):
-
     retval = con.run("SELECT :v", v=test_input, types={"v": oid})
     assert retval[0][0] == test_input
 
@@ -593,19 +555,71 @@ def test_roundtrip_oid(con, test_input, oid):
 
 
 @pytest.mark.parametrize(
-    "test_input,typ",
+    "test_input,typ,req_ver",
     [
-        [[True, False, None], "bool[]"],
-        [[IPv4Address("192.168.0.1")], "inet[]"],
-        [[Date(2021, 3, 1)], "date[]"],
-        [[Datetime(2001, 2, 3, 4, 5, 6)], "timestamp[]"],
-        [[Datetime(2001, 2, 3, 4, 5, 6, 0, Timezone.utc)], "timestamptz[]"],
-        [[Time(4, 5, 6)], "time[]"],
-        [[Timedelta(seconds=30)], "interval[]"],
-        [[{"name": "Apollo 11 Cave", "zebra": True, "age": 26.003}], "jsonb[]"],
-        [[b"\x00\x01\x02\x03\x02\x01\x00"], "bytea[]"],
-        [[Decimal("1.1"), None, Decimal("3.3")], "numeric[]"],
-        [[UUID("911460f2-1f43-fea2-3e2c-e01fd5b5069d")], "uuid[]"],
+        [[True, False, None], "bool[]", None],
+        [
+            [
+                Range(Date(2023, 6, 1), Date(2023, 6, 6)),
+                Range(Date(2023, 6, 10), Date(2023, 6, 13)),
+            ],
+            "datemultirange",
+            14,
+        ],
+        [Range(Date(1937, 6, 1), Date(2023, 5, 10)), "daterange", None],
+        ['"a"=>"1"', "hstore", None],
+        [[IPv4Address("192.168.0.1")], "inet[]", None],
+        [[Range(3, 7), Range(8, 9)], "int4multirange", 14],
+        [Range(3, 7), "int4range", None],
+        [50000000000000, "int8", None],
+        [[Range(3, 7), Range(8, 9)], "int8multirange", 14],
+        [Range(3, 7), "int8range", None],
+        [Range(3, 7), "numrange", None],
+        [[Range(3, 7), Range(Decimal("9.5"), Decimal("11.4"))], "nummultirange", 14],
+        [[Date(2021, 3, 1)], "date[]", None],
+        [[Datetime(2001, 2, 3, 4, 5, 6)], "timestamp[]", None],
+        [[Datetime(2001, 2, 3, 4, 5, 6, 0, Timezone.utc)], "timestamptz[]", None],
+        [[Time(4, 5, 6)], "time[]", None],
+        [
+            [
+                Range(Datetime(2001, 2, 3, 4, 5), Datetime(2023, 2, 3, 4, 5)),
+                Range(Datetime(2024, 6, 1), Datetime(2024, 7, 3)),
+            ],
+            "tsmultirange",
+            14,
+        ],
+        [
+            Range(Datetime(2001, 2, 3, 4, 5), Datetime(2023, 2, 3, 4, 5)),
+            "tsrange",
+            None,
+        ],
+        [
+            [
+                Range(
+                    Datetime(2001, 2, 3, 4, 5, tzinfo=Timezone.utc),
+                    Datetime(2023, 2, 3, 4, 5, tzinfo=Timezone.utc),
+                ),
+                Range(
+                    Datetime(2024, 6, 1, tzinfo=Timezone.utc),
+                    Datetime(2024, 7, 3, tzinfo=Timezone.utc),
+                ),
+            ],
+            "tstzmultirange",
+            14,
+        ],
+        [
+            Range(
+                Datetime(2001, 2, 3, 4, 5, tzinfo=Timezone.utc),
+                Datetime(2023, 2, 3, 4, 5, tzinfo=Timezone.utc),
+            ),
+            "tstzrange",
+            None,
+        ],
+        [[Timedelta(seconds=30)], "interval[]", None],
+        [[{"name": "Apollo 11 Cave", "zebra": True, "age": 26.003}], "jsonb[]", None],
+        [[b"\x00\x01\x02\x03\x02\x01\x00"], "bytea[]", None],
+        [[Decimal("1.1"), None, Decimal("3.3")], "numeric[]", None],
+        [[UUID("911460f2-1f43-fea2-3e2c-e01fd5b5069d")], "uuid[]", None],
         [
             [
                 "Hello!",
@@ -621,30 +635,32 @@ def test_roundtrip_oid(con, test_input, oid):
                 None,
             ],
             "varchar[]",
+            None,
         ],
-        [Timedelta(seconds=30), "interval"],
-        [Time(4, 5, 6), "time"],
-        [Date(2001, 2, 3), "date"],
-        ["infinity", "date"],
-        [Datetime(2001, 2, 3, 4, 5, 6), "timestamp"],
-        [Datetime(2001, 2, 3, 4, 5, 6, 0, Timezone.utc), "timestamptz"],
-        [True, "bool"],
-        [Decimal("1.1"), "numeric"],
-        [1.756e-12, "float8"],
-        [float("inf"), "float8"],
-        ["hello world", "unknown"],
-        ["hello \u0173 world", "varchar"],
-        [50000000000000, "int8"],
-        [b"\x00\x01\x02\x03\x02\x01\x00", "bytea"],
-        [bytearray(b"\x00\x01\x02\x03\x02\x01\x00"), "bytea"],
-        [UUID("911460f2-1f43-fea2-3e2c-e01fd5b5069d"), "uuid"],
-        [IPv4Network("192.168.0.0/28"), "inet"],
-        [IPv4Address("192.168.0.1"), "inet"],
+        [[], "varchar[]", None],
+        [Time(4, 5, 6), "time", None],
+        [Date(2001, 2, 3), "date", None],
+        ["infinity", "date", None],
+        [Datetime(2001, 2, 3, 4, 5, 6), "timestamp", None],
+        [Datetime(2001, 2, 3, 4, 5, 6, 0, Timezone.utc), "timestamptz", None],
+        [True, "bool", None],
+        [Decimal("1.1"), "numeric", None],
+        [1.756e-12, "float8", None],
+        [float("inf"), "float8", None],
+        ["hello world", "unknown", None],
+        ["hello \u0173 world", "varchar", None],
+        [50000000000000, "int8", None],
+        [b"\x00\x01\x02\x03\x02\x01\x00", "bytea", None],
+        [bytearray(b"\x00\x01\x02\x03\x02\x01\x00"), "bytea", None],
+        [UUID("911460f2-1f43-fea2-3e2c-e01fd5b5069d"), "uuid", None],
+        [IPv4Network("192.168.0.0/28"), "inet", None],
+        [IPv4Address("192.168.0.1"), "inet", None],
     ],
 )
-def test_roundtrip_cast(con, test_input, typ):
-    retval = con.run(f"SELECT CAST(:v AS {typ})", v=test_input)
-    assert retval[0][0] == test_input
+def test_roundtrip_cast(con, pg_version, test_input, typ, req_ver):
+    if req_ver is None or pg_version >= req_ver:
+        retval = con.run(f"SELECT CAST(:v AS {typ})", v=test_input)
+        assert retval[0][0] == test_input
 
 
 @pytest.mark.parametrize(
@@ -664,17 +680,6 @@ def test_array_in(con, test_input, expected):
     assert result[0][0] == expected
 
 
-def test_numeric_array_out(con):
-    res = con.run("SELECT '{1.1,2.2,3.3}'::numeric[] AS f1")
-    assert res[0][0] == [Decimal("1.1"), Decimal("2.2"), Decimal("3.3")]
-
-
-def test_empty_array(con):
-    v = []
-    retval = con.run("SELECT cast(:v as varchar[])", v=v)
-    assert retval[0][0] == v
-
-
 def test_macaddr(con):
     retval = con.run("SELECT macaddr '08002b:010203'")
     assert retval[0][0] == "08:00:2b:01:02:03"
@@ -687,12 +692,6 @@ def test_tsvector_roundtrip(con):
     assert retval[0][0] == "'a' 'and' 'ate' 'cat' 'fat' 'mat' 'on' 'rat' 'sat'"
 
 
-def test_hstore_roundtrip(con):
-    val = '"a"=>"1"'
-    retval = con.run("SELECT cast(:val as hstore)", val=val)
-    assert retval[0][0] == val
-
-
 def test_json_access_object(con):
     val = {"name": "Apollo 11 Cave", "zebra": True, "age": 26.003}
     retval = con.run("SELECT cast(:val as json) -> :name", val=dumps(val), name="name")
@@ -730,6 +729,31 @@ def test_jsonb_access_path(con):
     assert retval[0][0] == str(j[path[0]][int(path[1])])
 
 
-def test_time_in():
-    actual = time_in("12:57:18.000396")
-    assert actual == Time(12, 57, 18, 396)
+@pytest.fixture
+def duple_type(con):
+    con.run("CREATE TYPE duple AS (a int, b int);")
+    yield
+    con.run("DROP TYPE IF EXISTS duple;")
+
+
+@pytest.mark.parametrize(
+    "test_input,expected",
+    [
+        ((1, 3), "(1,3)"),
+        ((1, None), "(1,)"),
+    ],
+)
+def test_composite_type(con, duple_type, test_input, expected):
+    retval = con.run("SELECT CAST(:v AS duple)", v=test_input)
+    assert retval[0][0] == expected
+
+
+@pytest.mark.parametrize(
+    "test_input,expected",
+    [
+        ([(1, 3)], '{"(1,3)"}'),
+    ],
+)
+def test_composite_type_array(con, duple_type, test_input, expected):
+    retval = con.run("SELECT CAST(:v AS duple[])", v=test_input)
+    assert retval[0][0] == expected
